//
// Function for recursive calculation of Bessel and Neumann functions
// version 28.1.2002
// (c) 2002 Ingo D. Rullhusen <d01c@uni-bremen.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
//

#include <complex.h>
#include <math.h>
#include <float.h>
#include <iostream.h>


void cjn( int n, complex<double> x, complex<double> *J, complex<double> *Js )
{
   int nmax, i, k;
   complex<double> *j, *j_alloc, norm, jl, ju;

#ifdef BOUNDARY_CHECK
   if ( n < 0 )
   {
      cerr << "Error in cjn(): n<0!" << endl;
      exit(1);
   }
#endif

   jl = complex<double>(0.0,0.0);
   ju = complex<double>(0.0,0.0);

   for ( nmax=n+(int)ceil(abs(x))+3; ; nmax+=12 )
   {
      j = j_alloc = new complex<double>[nmax+2];
      j += 1;

      if ( abs(x) < DBL_EPSILON )
      {
         for ( i=-1; i<nmax; i++ ) j[i] = complex<double>(0.0,0.0);
         j[0] = complex<double>(1.0,0.0);
         break;
      }

      j[nmax] = complex<double>(0.0,0.0);
      j[nmax-1] = complex<double>(1.0,0.0);

      for ( i=nmax-2; i>-2; i-- )
      {
         norm = j[i] = ( 2.0*(i+1)/x ) * j[i+1] - j[i+2];
         if ( abs(norm) > 10.0 )
         {
            for ( k=nmax; k>i; k-- ) j[k] /= norm;
            j[i] = complex<double>(1.0,0.0);
         }
      }

      norm = j[0];
      for ( i=2; i<nmax; i+=2 ) norm += 2.0 * j[i];

      for ( i=-1; i<n+2; i++ ) j[i] /= norm;

      if ( abs( jl - j[-1]  ) < DBL_EPSILON &&
           abs( ju - j[n+1] ) < DBL_EPSILON    ) break;

      jl = j[-1];
      ju = j[n+1];

      delete [] j_alloc;
   }

   for ( i=0; i<=n; i++ )
   {
      J[i] = j[i];
      Js[i] = 0.5 * ( j[i-1] - j[i+1] );
   }

   delete [] j_alloc;

}


void cyn( int n, complex<double> x, complex<double> *Y, complex<double> *Ys )
{
   const double gamma = 0.577215664901532860606512;
   int nmin, k;
   complex<double> *y, *y_alloc, *j, *js, y0, y0o, y1, y1o;

#ifdef BOUNDARY_CHECK
   if ( n < 0 )
   {
      cerr << "Error in cyn(): n<0!" << endl;
      exit(1);
   }
#endif

   y0o = complex<double>(0.0,0.0);
   y1o = complex<double>(0.0,0.0);
   for ( nmin=n+1; ; nmin+=12 )
   {
      j = new complex<double>[nmin+1];
      js = new complex<double>[nmin+1];

      cjn( nmin, x, j, js );

      y0 = j[0] * ( log(0.5*x) + gamma ) * 2.0 / M_PI;
      y1 = - j[0] / ( M_PI*0.5*x )
           + j[1] * ( log(0.5*x) + gamma - 1.0 ) * 2.0 / M_PI;
      for ( k=1; k<(nmin-1)/2; k++ )
      {
         y0 -= 4.0 * pow( -1.0, k ) * j[2*k] / ( M_PI * k );
         y1 -= 2.0 * pow( -1.0, k ) * (1+2*k) * j[1+2*k] / ( M_PI*k*(1+k) );
      }

      if ( ( abs( y0 - y0o ) < DBL_EPSILON && abs( y1 - y1o ) < DBL_EPSILON )
           || abs( x ) < DBL_EPSILON ) break;

      y0o = y0;
      y1o = y1;

      delete [] j;
      delete [] js;
   }

   y = y_alloc = new complex<double>[n+3];
   y += 1;

   y[0] = y0;
   y[-1] = -y1;

   delete [] j;
   delete [] js;

   for ( k=1; k<n+2; k++ )
   {
      y[k] = ( 2.0*(k-1)/x ) * y[k-1] - y[k-2];
   }

   for ( k=0; k<=n; k++ )
   {
      Y[k] = y[k];
      Ys[k] = 0.5 * ( y[k-1] - y[k+1] );
   }

   delete [] y_alloc;

}
