//
// Class to compute and store main data
// version 1.2.2002
// (c) 2002 Ingo D. Rullhusen <d01c@uni-bremen.de>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
//

#include <complex.h>
#include <fstream.h>
#include <math.h>

#include "main_data.h"
#include "natconst.h"
#include "cjy.h"


static const complex<double> IU = complex<double>( 0.0, 1.0 );
static const complex<double> CZ = complex<double>( 0.0, 0.0 );

static inline complex<double> squ( complex<double> x ){ return(x*x); }


MainData::MainData( InputData idata )
{
   int n;
   complex<double> my1, *H, *Hs, *Ju, *Jsu, *Jv, *Jsv,
                   *H_alloc, *Hs_alloc, *Ju_alloc, *Jsu_alloc,
                   *Jv_alloc, *Jsv_alloc, Nq, K, u, v,
                   *jtmp, *jstmp, *ytmp, *ystmp, A, B, C, D, E;

   nmax = idata.max_n;
   omega = 2.0*M_PI*idata.freq;
   sinth = sin(idata.theta_inc);
   costh = cos(idata.theta_inc);
   my1 = MY0 * idata.my_r_1;
   my2 = MY0 * idata.my_r_2;
   beta1 = omega * sqrt( idata.eps_r_1 * idata.my_r_1 ) / C0;
   beta2 = omega * sqrt( idata.eps_r_2 * idata.my_r_2 ) / C0;

   asn = asn_alloc = new complex<double>[2*idata.max_n+1];
   asn += idata.max_n;
   bsn = bsn_alloc = new complex<double>[2*idata.max_n+1];
   bsn += idata.max_n;

   an = an_alloc = new complex<double>[2*idata.max_n+1];
   an += idata.max_n;
   bn = bn_alloc = new complex<double>[2*idata.max_n+1];
   bn += idata.max_n;

   H = H_alloc = new complex<double>[2*idata.max_n+1];
   H += idata.max_n;
   Hs = Hs_alloc = new complex<double>[2*idata.max_n+1];
   Hs += idata.max_n;
   Ju = Ju_alloc = new complex<double>[2*idata.max_n+1];
   Ju += idata.max_n;
   Jsu = Jsu_alloc = new complex<double>[2*idata.max_n+1];
   Jsu += idata.max_n;
   Jv = Jv_alloc = new complex<double>[2*idata.max_n+1];
   Jv += idata.max_n;
   Jsv = Jsv_alloc = new complex<double>[2*idata.max_n+1];
   Jsv += idata.max_n;

   Nq = ( idata.eps_r_1 * idata.my_r_1 ) / ( idata.eps_r_2 * idata.my_r_2 );

   K = idata.my_r_1 / idata.my_r_2;

   u = sqrt( squ(beta1) - squ(beta2*costh) ) * idata.a;

   v = beta2*sinth*idata.a;

   jtmp = new complex<double>[idata.max_n+1];
   jstmp = new complex<double>[idata.max_n+1];
   cjn( idata.max_n, v, jtmp, jstmp );

   ytmp = new complex<double>[idata.max_n+1];    
   ystmp = new complex<double>[idata.max_n+1];
   cyn( idata.max_n, v, ytmp, ystmp );

   if ( abs(jtmp[0]) < sqrt(DBL_MIN) || abs(jstmp[0]) < sqrt(DBL_MIN) )
   {
      cerr << "Error: Bessel function underflow!" << endl;
      exit(1);
   }
   if ( abs(ytmp[0]) < sqrt(DBL_MIN) || abs(ystmp[0]) < sqrt(DBL_MIN) )
   {
      cerr << "Error: Neumann function underflow!" << endl;
      exit(1);
   }
   Jv[0] = jtmp[0];
   Jsv[0] = jstmp[0];
   H[0] = jtmp[0] - IU * ytmp[0];
   Hs[0] = jstmp[0] - IU * ystmp[0];
   for ( n=1; n<=idata.max_n; n++ )
   {
      if ( abs(jtmp[n]) < sqrt(DBL_MIN) || abs(jstmp[n]) < sqrt(DBL_MIN) )
      {
         cerr << "Error: Bessel function underflow!" << endl;
         exit(1);
      }
      if ( abs(ytmp[n]) < sqrt(DBL_MIN) || abs(ystmp[n]) < sqrt(DBL_MIN) )
      {
         cerr << "Error: Neumann function underflow!" << endl;
         exit(1);
      }
      Jv[n] = jtmp[n];
      Jsv[n] = jstmp[n];
      H[n] = jtmp[n] - IU * ytmp[n];
      Hs[n] = jstmp[n] - IU * ystmp[n];

      Jv[-n] = pow( -1.0, n ) * Jv[n];
      Jsv[-n] = pow( -1.0, n ) * Jsv[n];
      H[-n] = pow( -1.0, n ) * H[n];
      Hs[-n] = pow( -1.0, n ) * Hs[n];
   }

   cjn( idata.max_n, u, jtmp, jstmp );

   Ju[0] = jtmp[0];
   Jsu[0] = jstmp[0];
   for ( n=1; n<=idata.max_n; n++ )
   {
      if ( abs(jtmp[n]) < sqrt(DBL_MIN) || abs(jstmp[n]) < sqrt(DBL_MIN) )
      {
         cerr << "Error: Bessel function underflow!" << endl;
         exit(1);
      }
      Ju[n] = jtmp[n];
      Jsu[n] = jstmp[n];
      Ju[-n] = pow( -1.0, n ) * jtmp[n];
      Jsu[-n] = pow( -1.0, n ) * jstmp[n];

   }

   delete [] ystmp;
   delete [] ytmp;
   delete [] jstmp;
   delete [] jtmp;

   for ( n=-idata.max_n; n<=idata.max_n; n++ )
   {

/* normal incidence only */
/*
      asn[n] = - idata.E_ampl * pow( IU, n )
               * ( ( Nq*Jsu[n]/(K*u*Ju[n]) - Jsv[n]/(v*Jv[n]) )
                   / ( Nq*Jsu[n]/(K*u*Ju[n]) - Hs[n]/(v*H[n]) ) )
               * Jv[n]/H[n];

      bsn[n] = CZ;
*/

      A = n*beta2*costh * ( 1.0/squ(u) - 1.0/squ(v) ) * H[n];

      B = ( IU*my2*omega*(Hs[n]/(v*H[n])) - IU*my1*omega*(Jsu[n]/(u*Ju[n])) )
          * H[n];

      C = idata.E_ampl * pow( IU, n ) * sinth * n*beta2*costh
          * ( 1.0/squ(v) - 1.0/squ(u) ) * Jv[n];

      D = ( IU*squ(beta1)/(my1*omega) * (Jsu[n]/(u*Ju[n]))
            - IU*squ(beta2)/(my2*omega) * (Hs[n]/(v*H[n])) ) * H[n];

      E = idata.E_ampl * pow( IU, n ) * sinth
          * ( IU*squ(beta2)/(my2*omega) * (Jsv[n]/(v*Jv[n]))
              - IU*squ(beta1)/(my1*omega) * (Jsu[n]/(u*Ju[n])) ) * Jv[n];

      asn[n] = ( A*C - B*E ) / ( squ(A) - B*D );
      bsn[n] = ( A*E - C*D ) / ( squ(A) - B*D );

      an[n] = ( idata.E_ampl * pow( IU, n ) * sinth * Jv[n]
                + asn[n] * H[n] ) / Ju[n];
      bn[n] = bsn[n] * H[n] / Ju[n];

   }

   delete [] Jsv_alloc;
   delete [] Jv_alloc;
   delete [] Jsu_alloc;
   delete [] Ju_alloc;
   delete [] Hs_alloc;
   delete [] H_alloc;

}


MainData::~MainData()
{
   delete [] bsn_alloc;
   delete [] asn_alloc;
   delete [] bn_alloc;
   delete [] an_alloc;
}


void MainData::ffout( InputData idata )
{
   ofstream ofile;
   int p, n;
   double phi, RCSp, RCSr, RCSz;
   complex<double> Ep, Er, Ez;

   ofile.open( idata.ff_file, ios::out );
   ofile << "# RCS per wavelength over azimuth" << endl;
   ofile << "# phi/ & RCS_phi (lin) & RCS_rho (lin) & RCS_z (lin) ";
   ofile << "& RCS_phi (dB) & RCS_rho (dB) & RCS_z (dB)" << endl;

   for ( p=0; p<idata.ffno; p++ )
   {
      phi = p*M_PI/(idata.ffno-1);

// z=0
      Ep = CZ;
      Er = CZ;
      Ez = CZ;
      for ( n=-idata.max_n; n<=idata.max_n; n++ )
      {
         Ez += asn[n] * pow( IU, n ) * exp( -IU*n*phi );
         Ep += bsn[n] * pow( IU, n ) * exp( -IU*n*phi );
      }
// sqrt(2*pi*rho) included
      Ez *= + sqrt( 4.0*IU/(beta2*sinth) );
      Ep *= - sqrt( 4.0*IU/(beta2*sinth) ) * ( my2*omega/(beta2*sinth) );
      Er = - Ez * costh / sinth;

      RCSz = abs( Ez*conj(Ez) ) / abs( idata.E_ampl*conj(idata.E_ampl) );
      RCSz /= 2.0*M_PI/real(beta2);

      RCSp = abs( Ep*conj(Ep) ) / abs( idata.E_ampl*conj(idata.E_ampl) );
      RCSp /= 2.0*M_PI/real(beta2);

      RCSr = abs( Er*conj(Er) ) / abs( idata.E_ampl*conj(idata.E_ampl) );
      RCSr /= 2.0*M_PI/real(beta2);

      ofile << 180.0*phi/M_PI << " ";
      ofile << RCSp << " " << RCSr << " " << RCSz << " ";

      if ( RCSp > DBL_EPSILON )
         ofile << 10.0*log10(RCSp) << " ";
      else
         ofile << 10.0*log10(DBL_EPSILON) << " ";

      if ( RCSr > DBL_EPSILON )
         ofile << 10.0*log10(RCSr) << " ";
      else
         ofile << 10.0*log10(DBL_EPSILON) << " ";

      if ( RCSz > DBL_EPSILON )
         ofile << 10.0*log10(RCSz) << " ";
      else
         ofile << 10.0*log10(DBL_EPSILON) << " ";

      ofile << endl;
   }

   ofile.close();

}


complex<double> MainData::Eout( InputData idata )
{
   int n;
   complex<double> *J, *J_alloc, *jtmp, *jstmp, x, E;

   jtmp = new complex<double>[nmax+1];
   jstmp = new complex<double>[nmax+1];

   x = sqrt( squ(beta1) - squ(beta2*costh) ) * idata.rho;

   cjn( nmax, x, jtmp, jstmp );

   J = J_alloc = new complex<double>[2*nmax+1];
   J += nmax;

   J[0] = jtmp[0];
   for ( n=1; n<=nmax; n++ )
   {
      J[n] = jtmp[n];
      J[-n] = pow( -1.0, n ) * J[n];
   }

   E = CZ;
   for ( n=-nmax; n<=nmax; n++ )
   {
      E += an[n] * J[n] * exp( -IU*n*idata.phi -IU*beta2*idata.z*costh );
   }

   delete [] jtmp;
   delete [] jstmp;
   delete [] J_alloc;

   return(E);
}
